/*****************************************************************************
        Author:         Stephen Strudwick
        Department:     Unix Programming Group
        Modify Date:    09/01/2004
        Company:        Pipex Communications

	Description:	API for blowfish

					This is basically an API for the blowfish code.

******************************************************************************/

#include "bf_crypt.h"

/*****************************************************************************
   Description: Initialise a Blowfish key.
   Inputs:      char *key
   Output:      0 on success
                Negative on error
******************************************************************************/

BFkey_type *blowfish_init(char *key)
{
  int key_len;
  BFkey_type * ks;
  ks = (BFkey_type *)Malloc(8192);

  if (key)
    key_len = strlen(key);
  else
    return NULL;

  if (key_len <8 || key_len > 56)
    return NULL;

  if (blowfish_make_bfkey(key, key_len, ks) != 0 ) {
    return NULL;
  }
  return ks;
}

/*****************************************************************************
   Description: Destroys a Blowfish key.
   Inputs:      void
   Output:      void
******************************************************************************/

void blowfish_destroy(BFkey_type *ks)
{
  free(ks);
}

/*****************************************************************************
   Description: Encrypts a string of any length, the encrypted result is always
                mod 8. If the input is not mod 8 it is padded to that length
                before encryption.

                Input is char* but output is a struct containing char* and the
                length of the encrypted string, this is done because it is
                possible to have NULL characters in the encrypted string, hence
                we cant just use strlen to work out length of string.

   Inputs:      char *input, c_string *output
   Output:      void
******************************************************************************/

void blowfish_crypt(BFkey_type *ks, char *input, c_string *output, int len) {
  char *ptr, buf[9];
  int n;

  if (!input)
    return;

  if ((len % 8) != 0) {
    output->len = (floor(len / 8)+1)*8;
  } else {
    output->len = len;
  }

  output->c_string = (char *)Malloc(output->len+1);
  ptr = output->c_string;

  for (n=0;n<(output->len / 8);n++) {
    memset(buf, '\0', 9);	/* Set all mem space to NULL */
    memcpy(buf, input+(n*8), 8); /* copy 8 bytes from buf into space */
    blowfish_crypt_8bytes((unsigned char*)buf, (unsigned char*)ptr, ks, 0);
    ptr = &output->c_string[(n+1)*8];
  }
  *ptr = NUL;
}

/*****************************************************************************
   Description: Decrypts an encrypted string. The input must be mod 8 of the
                function fails and returns NULL.
   Inputs:      char *input
   Output:      char *
******************************************************************************/

char *blowfish_decrypt(BFkey_type *ks, char *input, int len) {
  int n;
  char *ptr, *output, buf[9];

  if ((len % 8) != 0) {
    return NULL;
  }

  output = (char *)Malloc(len+1);
  ptr = output;

  for (n=0;n<(len / 8);n++) {
    memcpy(buf, input+(n*8), 8);
    buf[8] = NUL;
    blowfish_crypt_8bytes((unsigned char*)buf, (unsigned char*)ptr, ks, 1);
    ptr = &output[(n+1)*8];
  }
  *ptr = NUL;

  return output;
}
