/******************************************************************************\
* wincompat.c
*
* Some compabillity routines to help VC compile the unix sources as much as possible..
* These are only tested with the nrpe sources, and may break on anything else...
*
* NRPE_NT: Nagios Remote Plugin Executor for NT/W2K 
* Copyright (c) 2003 Michael Wirtgen (michael.wirtgen@miwi-dv.com)
* 
* based on:
* NRPE.C - Nagios Remote Plugin Executor
* Copyright (c) 1999-2003 Ethan Galstad (nagios@nagios.org)
*
* License: GPL
* 
* Nagios and the Nagios logo are registered trademarks of Ethan Galstad.
*
* strsep code taken from libgw32c-0.3 (http://gnuwin32.sourceforge.net)
* License: GPL
*
* syslog code modified from wsyslog.c from the php sources with the following
* copyright notice:
*
* Copyright 1996 by the University of Washington
*
*  Permission to use, copy, modify, and distribute this software and its
* documentation for any purpose and without fee is hereby granted, provided
* that the above copyright notice appears in all copies and that both the
* above copyright notice and this permission notice appear in supporting
* documentation, and that the name of the University of Washington not be
* used in advertising or publicity pertaining to distribution of the software
* without specific, written prior permission.  This software is made available
* "as is", and
* THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
* WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT LIMITATION ALL IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND IN
* NO EVENT SHALL THE UNIVERSITY OF WASHINGTON BE LIABLE FOR ANY SPECIAL,
* INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
* LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT
* (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION
* WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*
*
\******************************************************************************/

#include <stdio.h>
#include <conio.h>
#include <string.h>
#include <windows.h>
#include "wincompat.h"
#include "..\evlog.h"

const char *EVENTLOG_KEY = "SYSTEM\\CurrentControlSet\\Services\\EventLog\\Application\\NRPE_NT"; 
	

void syslog(int priority, const char *message, ...)
{
	va_list args;
	LPTSTR strs[2];
    char szEvent[4096];	
	char tmp[4096];				/* callers must be careful not to pop this */
	unsigned short etype;
	DWORD eID;
    HANDLE loghdl = NULL;

	
	/* default event source */
		
	switch (priority) {			/* translate UNIX type into NT type */
		case LOG_ALERT:
			etype = EVENTLOG_ERROR_TYPE;
			eID = EVMSGERROR;
			break;
		case LOG_INFO:
			etype = EVENTLOG_INFORMATION_TYPE;
			eID = EVMSGINFO;
			break;
		case LOG_DEBUG:
			etype = EVENTLOG_INFORMATION_TYPE;
			eID = EVMSGINFO;
			break;

		default:
			etype = EVENTLOG_WARNING_TYPE;
			eID = EVMSGWARNING;
	}

	// Open handle to eventlog
    loghdl = RegisterEventSource(NULL, "NRPE_NT");
    
	va_start(args, message);	/* initialize vararg mechanism */
	vsprintf(tmp, message, args);	/* build message */
	wsprintf(szEvent, "%s: ", "NRPE_NT");
    strs[0]=szEvent;
    strs[1]=tmp;
    /* report the event */
	ReportEvent(loghdl, etype, (unsigned short) priority, eID, NULL, 2, 0, strs, NULL);
	va_end(args);

    // Close evntlog handle
    DeregisterEventSource(loghdl);
	return;
}


// stub for bzero unix call
void bzero(char * dest, int length)
{
ZeroMemory( dest, length);
}

// Sleep takes ms instead of seconds on win
void sleep (int seconds)
{
Sleep(100*seconds);
}

int InstallService(char * ExePath)
{
	SC_HANDLE srv,scm;
	HKEY  hKey;
	DWORD dwtemp;
	scm=OpenSCManager(0,0,SC_MANAGER_CREATE_SERVICE);
	if(!scm)
	{
		return 1;
	}
	srv=CreateService(scm,"NRPE_NT",
		"Nagios Remote Plugin Executor for NT/W2K",
		SERVICE_ALL_ACCESS,SERVICE_WIN32_OWN_PROCESS,SERVICE_AUTO_START,
		SERVICE_ERROR_NORMAL,
		ExePath,
		0,0,0,0,0);
	if(!srv)
	{
		CloseServiceHandle(scm);
		return 1;
	}
	CloseServiceHandle(srv);
	CloseServiceHandle(scm);

	if (RegCreateKeyEx(HKEY_LOCAL_MACHINE, EVENTLOG_KEY, 0, NULL, 
					 REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hKey, &dwtemp) 
					  == ERROR_SUCCESS)
	  {
		RegSetValueEx(hKey, "EventMessageFile", 0, REG_EXPAND_SZ,(LPBYTE)ExePath,
						  strlen(ExePath) + 1);
		dwtemp = EVENTLOG_ERROR_TYPE | EVENTLOG_WARNING_TYPE | EVENTLOG_INFORMATION_TYPE;
		RegSetValueEx(hKey, "TypesSupported", 0, REG_DWORD,(LPBYTE)&dwtemp, sizeof(DWORD));
		RegCloseKey(hKey);
	  }
	else
	{
		printf("Could not register EventSource!");
		syslog(LOG_ERR,"Could not register EventSource!");
	}

	return 0;
}

int UninstallService()
{
	SC_HANDLE srv,scm;
    int res = 0;

	scm=OpenSCManager(0,0,SC_MANAGER_CREATE_SERVICE);
	if(!scm)
	{
		return 1;
	}
	srv=OpenService(scm,"NRPE_NT",DELETE);
	if(!srv)
	{
		CloseServiceHandle(scm);
		return 1;
	}
	if (!DeleteService(srv))
        res = 1;

	if(!(RegDeleteKey(HKEY_LOCAL_MACHINE, EVENTLOG_KEY)==ERROR_SUCCESS))
	{
		printf("Could not unregister EventSource!");
		syslog(LOG_ERR,"Could not unregister EventSource!");
	}

    
    CloseServiceHandle(srv);
	CloseServiceHandle(scm);
	return res;
}